/**
 * @file cstrlist.c
 * @brief Zero separated string list.
 *
 * Copyright 2022 Raymond Czerny
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */


#include <iso646.h> // boolean key words: not, and, or, xor ...
#include <stdlib.h>
#include <string.h>

#include "cstrlist.h"

#ifdef _WIN32
#define strcasecmp _stricmp
#define wcscasecmp _wcsicmp
#endif

// === String list with Single-Byte-Characters ===========

bool cstrlist_create(CStrList* list, size_t count)
{
    list->begin = calloc(count, sizeof(char));
    list->end = list->begin + count;
    return (NULL == list->begin) ? false : true;
}

void cstrlist_destroy(CStrList* list)
{
    free(list->begin);
    list->begin = NULL;
    list->end = NULL;
}

void cstrlist_clean(CStrList* list)
{
    size_t size = list->end - list->begin;
    memset(list->begin, 0, size);
}

bool cstrlist_append(CStrList* list, const char* item)
{
    char* pos = list->begin;
    if((NULL == pos) or (NULL == item))
    {
        return false;
    }

    while(*pos and (pos < list->end))
    {
        pos += strlen(pos) + 1;
    }
    if((pos + strlen(item) + 2) < list->end)
    {
        strcpy(pos, item);
        return true;
    }
    else // more memory
    {
        size_t offset = pos - list->begin;
        size_t oldsize = list->end - list->begin;
        size_t delta = strlen(item) + 2;
        list->begin = realloc(list->begin, oldsize + delta);
        if(list->begin)
        {
            list->end = list->begin + oldsize + delta;
            memset(list->begin + offset, 0, delta);
            strcpy(list->begin + offset, item);
            return true;
        }
        list->end = NULL;
    }
    return false;
}

bool cstrlist_append_unique(CStrList* list, const char* item)
{
    if(SIZE_MAX == cstrlist_index(list, item))
    {
        return cstrlist_append(list, item);
    }
    return false;
}

bool cstrlist_append_unique_i(CStrList* list, const char* item)
{
    if(SIZE_MAX == cstrlist_index_i(list, item))
    {
        return cstrlist_append(list, item);
    }
    return false;
}

bool cstrlist_remove(CStrList* list, size_t index)
{
    char* tmp = NULL;
    char* found = NULL;
    char* pos = list->begin;
    if(NULL == pos)
    {
        return false;
    }
    for(size_t idx = 0; (pos < list->end); ++idx)
    {
        if(idx == index)
        {
            found = pos;
            break;
        }
        pos += strlen(pos) + 1;
    }
    if(NULL == found)
    {
        return false;
    }
    pos += strlen(pos) + 1;

    // save rest list
    if(pos < list->end)
    {
        size_t size = list->end - pos;
        tmp = calloc(size, sizeof(char));
        memcpy(tmp, pos, size);
    }

    // clear from found to end
    size_t size2 = list->end - found;
    memset(found, 0, size2);

    // add rest
    if(NULL != tmp)
    {
        memcpy(found, tmp, size2);
        free(tmp);
    }
    return true;
}

const char* cstrlist_value(CStrList* list, size_t index)
{
    char* pos = list->begin;
    if(NULL == pos)
    {
        return NULL;
    }
    for(size_t idx = 0; (pos < list->end); ++idx)
    {
        if(idx == index)
        {
            return pos;
        }
        pos += strlen(pos) + 1;
    }
    return NULL;
}

size_t cstrlist_count(CStrList* list)
{
    size_t count = 0;
    char* pos = list->begin;
    if(NULL == pos)
    {
        return count;
    }
    while((0 != *pos) and (pos < list->end))
    {
        pos += strlen(pos) + 1;
        ++count;
    }
    return count;
}

size_t cstrlist_index(CStrList* list, const char* item)
{
    size_t index = 0;
    char* pos = list->begin;
    if((NULL == pos) or (NULL == item))
    {
        return SIZE_MAX;
    }
    while((0 != *pos) and (pos < list->end))
    {
        if(strcmp(pos, item))
        {
            pos += strlen(pos) + 1;
            ++index;
        }
        else
        {
            return index;
        }
    }
    return SIZE_MAX;
}

size_t cstrlist_index_i(CStrList* list, const char* item)
{
    size_t index = 0;
    char* pos = list->begin;
    if((NULL == pos) or (NULL == item))
    {
        return SIZE_MAX;
    }
    while((0 != *pos) and (pos < list->end))
    {
        if(strcasecmp(pos, item))
        {
            pos += strlen(pos) + 1;
            ++index;
        }
        else
        {
            return index;
        }
    }
    return SIZE_MAX;
}

char* cstrlist_next(CStrList* list, char* item)
{
    if(NULL == item)
    {
        return list->begin;
    }
    if((item < list->begin) or (item >= list->end))
    {
        return NULL;
    }

    return item +  strlen(item) + 1;
}

// === String list with Wide-Characters ===========

bool wcstrlist_create(WCStrList* list, size_t count)
{
    list->begin = calloc(count, sizeof(wchar_t));
    list->end = list->begin + count;
    return (NULL == list->begin) ? false : true;

}

void wcstrlist_destroy(WCStrList* list)
{
    free(list->begin);
    list->begin = NULL;
    list->end = NULL;
}

void wcstrlist_clean(WCStrList* list)
{
    size_t size = list->end - list->begin;
    memset(list->begin, 0, size);
}

bool wcstrlist_append(WCStrList* list, const wchar_t* item)
{
    wchar_t* pos = list->begin;
    if((NULL == pos) or (NULL == item))
    {
        return false;
    }

    while((0 != *pos) and (pos < list->end))
    {
        pos += wcslen(pos) + 1;
    }
    if((pos + wcslen(item) + 1) < list->end)
    {
        wcscpy(pos, item);
        return true;
    }
    else // more memory
    {
        size_t offset = pos - list->begin;
        size_t oldsize = list->end - list->begin;
        size_t delta = wcslen(item) + 2;
        list->begin = realloc(list->begin, (oldsize + delta) * sizeof(wchar_t));
        if(list->begin)
        {
            list->end = list->begin + oldsize + delta;
            memset(list->begin + offset, 0, delta * sizeof(wchar_t));
            wcscpy(list->begin + offset, item);
            return true;
        }
        list->end = NULL;
    }
    return false;
}

bool wcstrlist_append_unique(WCStrList* list, const wchar_t* item)
{
    if(SIZE_MAX == wcstrlist_index(list, item))
    {
        return wcstrlist_append(list, item);
    }
    return false;
}

bool wcstrlist_append_unique_i(WCStrList* list, const wchar_t* item)
{
    if(SIZE_MAX == wcstrlist_index_i(list, item))
    {
        return wcstrlist_append(list, item);
    }
    return false;
}

bool wcstrlist_remove(WCStrList* list, size_t index)
{
    wchar_t* tmp = NULL;
    wchar_t* found = NULL;
    wchar_t* pos = list->begin;
    if(NULL == pos)
    {
        return false;
    }
    for(size_t idx = 0; (pos < list->end); ++idx)
    {
        if(idx == index)
        {
            found = pos;
            break;
        }
        pos += wcslen(pos) + 1;
    }
    if(NULL == found)
    {
        return false;
    }
    pos += wcslen(pos) + 1;

    // save rest list
    if(pos < list->end)
    {
        size_t size = list->end - pos;
        tmp = calloc(size, sizeof(wchar_t));
        memcpy(tmp, pos, size * sizeof(wchar_t));
    }

    // clear from found to end
    size_t size2 = list->end - found;
    memset(found, 0, size2 * sizeof(wchar_t));

    // add rest
    if(NULL != tmp)
    {
        memcpy(found, tmp, size2 * sizeof(wchar_t));
        free(tmp);
    }
    return true;
}

const wchar_t* wcstrlist_value(WCStrList* list, size_t index)
{
    wchar_t* pos = list->begin;
    if(NULL == pos)
    {
        return NULL;
    }
    for(size_t idx = 0; (pos < list->end); ++idx)
    {
        if(idx == index)
        {
            return pos;
        }
        pos += wcslen(pos) + 1;
    }
    return NULL;
}

size_t wcstrlist_count(WCStrList* list)
{
    size_t count = 0;
    wchar_t* pos = list->begin;
    if(NULL == pos)
    {
        return count;
    }
    while((0 != *pos) and (pos < list->end))
    {
        pos += wcslen(pos) + 1;
        ++count;
    }
    return count;
}

size_t wcstrlist_index(WCStrList* list, const wchar_t* item)
{
    size_t index = 0;
    wchar_t* pos = list->begin;
    if((NULL == pos) or (NULL == item))
    {
        return SIZE_MAX;
    }
    while((0 != *pos) and (pos < list->end))
    {
        if(wcscmp(pos, item))
        {
            pos += wcslen(pos) + 1;
            ++index;
        }
        else
        {
            return index;
        }
    }
    return SIZE_MAX;
}

size_t wcstrlist_index_i(WCStrList* list, const wchar_t* item)
{
    size_t index = 0;
    wchar_t* pos = list->begin;
    if((NULL == pos) or (NULL == item))
    {
        return SIZE_MAX;
    }
    while((0 != *pos) and (pos < list->end))
    {
        if(wcscasecmp(pos, item))
        {
            pos += wcslen(pos) + 1;
            ++index;
        }
        else
        {
            return index;
        }
    }
    return SIZE_MAX;
}

wchar_t* wcstrlist_next(WCStrList* list, wchar_t* item)
{
    if(NULL == item)
    {
        return list->begin;
    }
    if((item < list->begin) or (item >= list->end))
    {
        return NULL;
    }

    return item +  wcslen(item) + 1;
}
